// Copyright (c) .NET Foundation. All rights reserved.
// Licensed under the Apache License, Version 2.0. See License.txt in the project root for license information.

using System;

namespace Microsoft.EntityFrameworkCore.Metadata
{
    /// <summary>
    ///     <para>
    ///         Represents a scalar property of an entity.
    ///     </para>
    ///     <para>
    ///         This interface is used during model creation and allows the metadata to be modified.
    ///         Once the model is built, <see cref="IProperty" /> represents a ready-only view of the same metadata.
    ///     </para>
    /// </summary>
    public interface IMutableProperty : IProperty, IMutablePropertyBase
    {
        /// <summary>
        ///     Gets the type that this property belongs to.
        /// </summary>
        new IMutableEntityType DeclaringEntityType { get; }

        /// <summary>
        ///     Gets or sets a value indicating whether this property can contain null.
        /// </summary>
        new bool IsNullable { get; set; }

        /// <summary>
        ///     Gets or sets a value indicating when a value for this property will be generated by the database. Even when the
        ///     property is set to be generated by the database, EF may still attempt to save a specific value (rather than
        ///     having one generated by the database) when the entity is added and a value is assigned, or the property is
        ///     marked as modified for an existing entity. See <see cref="BeforeSaveBehavior" /> and <see cref="AfterSaveBehavior" />
        ///     for more information.
        /// </summary>
        new ValueGenerated ValueGenerated { get; set; }

        /// <summary>
        ///     <para>
        ///         Gets a value indicating whether or not this property can be modified before the entity is
        ///         saved to the database.
        ///     </para>
        ///     <para>
        ///         If <see cref="PropertySaveBehavior.Throw" />, then an exception
        ///         will be thrown if a value is assigned to this property when it is in
        ///         the <see cref="EntityState.Added" /> state.
        ///     </para>
        ///     <para>
        ///         If <see cref="PropertySaveBehavior.Ignore" />, then any value
        ///         set will be ignored when it is in the <see cref="EntityState.Added" /> state.
        ///     </para>
        /// </summary>
        new PropertySaveBehavior BeforeSaveBehavior { get; set; }

        /// <summary>
        ///     <para>
        ///         Gets a value indicating whether or not this property can be modified after the entity is
        ///         saved to the database.
        ///     </para>
        ///     <para>
        ///         If <see cref="PropertySaveBehavior.Throw" />, then an exception
        ///         will be thrown if a new value is assigned to this property after the entity exists in the database.
        ///     </para>
        ///     <para>
        ///         If <see cref="PropertySaveBehavior.Ignore" />, then any modification to the
        ///         property value of an entity that already exists in the database will be ignored.
        ///     </para>
        /// </summary>
        new PropertySaveBehavior AfterSaveBehavior { get; set; }

        /// <summary>
        ///     This property is obsolete. Use <see cref="BeforeSaveBehavior" /> instead.
        /// </summary>
        [Obsolete("Use BeforeSaveBehavior instead.")]
        new bool IsReadOnlyBeforeSave { get; set; }

        /// <summary>
        ///     This property is obsolete. Use <see cref="BeforeSaveBehavior" /> instead.
        /// </summary>
        [Obsolete("Use AfterSaveBehavior instead.")]
        new bool IsReadOnlyAfterSave { get; set; }

        /// <summary>
        ///     This property is obsolete. Use <see cref="BeforeSaveBehavior" /> or <see cref="AfterSaveBehavior" /> instead.
        /// </summary>
        [Obsolete("Use BeforeSaveBehavior or AfterSaveBehavior instead.")]
        new bool IsStoreGeneratedAlways { get; set; }

        /// <summary>
        ///     Gets or sets a value indicating whether this property is used as a concurrency token. When a property is configured
        ///     as a concurrency token the value in the database will be checked when an instance of this entity type
        ///     is updated or deleted during <see cref="DbContext.SaveChanges()" /> to ensure it has not changed since
        ///     the instance was retrieved from the database. If it has changed, an exception will be thrown and the
        ///     changes will not be applied to the database.
        /// </summary>
        new bool IsConcurrencyToken { get; set; }
    }
}
