// Copyright (c) .NET Foundation. All rights reserved.
// Licensed under the Apache License, Version 2.0. See License.txt in the project root for license information.

using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using System.Transactions;
using Microsoft.EntityFrameworkCore.Internal;
using Microsoft.EntityFrameworkCore.TestUtilities;
using Microsoft.Extensions.DependencyInjection;
using Xunit;

// ReSharper disable AccessToModifiedClosure
// ReSharper disable InconsistentNaming
namespace Microsoft.EntityFrameworkCore.Storage
{
    public class ExecutionStrategyTest : IDisposable
    {
        private readonly DbContext Context;

        public ExecutionStrategyTest() => Context = CreateContext();

        public void Dispose() => Context.Dispose();

        private TestExecutionStrategy CreateFailOnRetryStrategy()
            => new TestExecutionStrategy(
                Context,
                shouldRetryOn: e =>
                {
                    Assert.True(false);
                    return false;
                },
                getNextDelay: e =>
                {
                    Assert.True(false);
                    return null;
                });

        [Fact]
        public void GetNextDelay_returns_the_expected_default_sequence()
        {
            var strategy = new TestExecutionStrategy(Context);
            var delays = new List<TimeSpan>();
            var delay = strategy.GetNextDelayBase(new Exception());
            while (delay != null)
            {
                delays.Add(delay.Value);
                delay = strategy.GetNextDelayBase(new Exception());
            }

            var expectedDelays = new List<TimeSpan>
            {
                TimeSpan.FromSeconds(0),
                TimeSpan.FromSeconds(1),
                TimeSpan.FromSeconds(3),
                TimeSpan.FromSeconds(7),
                TimeSpan.FromSeconds(15),
                TimeSpan.FromSeconds(31)
            };

            Assert.Equal(expectedDelays.Count, delays.Count);
            for (var i = 0; i < expectedDelays.Count; i++)
            {
                Assert.True(
                    Math.Abs((delays[i] - expectedDelays[i]).TotalMilliseconds)
                    <= expectedDelays[i].TotalMilliseconds * 0.1 + 1,
                    string.Format("Expected: {0}; Actual: {1}", expectedDelays[i], delays[i]));
            }
        }

        [Fact]
        public void RetriesOnFailure_returns_true()
        {
            var mockExecutionStrategy = new TestExecutionStrategy(Context);

            Assert.True(mockExecutionStrategy.RetriesOnFailure);
        }

        [Fact]
        public void Execute_Action_throws_for_an_existing_transaction()
        {
            Execute_throws_for_an_existing_transaction(e => e.Execute(() => { }));
        }

        [Fact]
        public void Execute_Func_throws_for_an_existing_transaction()
        {
            Execute_throws_for_an_existing_transaction(e => e.Execute(() => 1));
        }

        private void Execute_throws_for_an_existing_transaction(Action<ExecutionStrategy> execute)
        {
            var mockExecutionStrategy = new TestExecutionStrategy(Context);
            using (Context.Database.BeginTransaction())
            {
                Assert.Equal(
                    CoreStrings.ExecutionStrategyExistingTransaction(mockExecutionStrategy.GetType().Name, "DbContext.Database.CreateExecutionStrategy()"),
                    Assert.Throws<InvalidOperationException>(
                            () => execute(mockExecutionStrategy))
                        .Message);
            }
        }

        [Fact]
        public void Execute_Action_throws_for_an_ambient_transaction()
        {
            Execute_throws_for_an_ambient_transaction(e => e.Execute(() => { }));
        }

        [Fact]
        public void Execute_Func_throws_for_an_ambient_transaction()
        {
            Execute_throws_for_an_ambient_transaction(e => e.Execute(() => 1));
        }

        private void Execute_throws_for_an_ambient_transaction(Action<ExecutionStrategy> execute)
        {
            var mockExecutionStrategy = new TestExecutionStrategy(Context);
            using (new TransactionScope())
            {
                Assert.Equal(
                    CoreStrings.ExecutionStrategyExistingTransaction(mockExecutionStrategy.GetType().Name, "DbContext.Database.CreateExecutionStrategy()"),
                    Assert.Throws<InvalidOperationException>(
                            () => execute(mockExecutionStrategy))
                        .Message);
            }
        }

        [Fact]
        public void Execute_Action_throws_for_an_enlisted_transaction()
        {
            Execute_throws_for_an_enlisted_transaction(e => e.Execute(() => { }));
        }

        [Fact]
        public void Execute_Func_throws_for_an_enlisted_transaction()
        {
            Execute_throws_for_an_enlisted_transaction(e => e.Execute(() => 1));
        }

        private void Execute_throws_for_an_enlisted_transaction(Action<ExecutionStrategy> execute)
        {
            var mockExecutionStrategy = new TestExecutionStrategy(Context);
            using (var t = new CommittableTransaction())
            {
                Context.Database.EnlistTransaction(t);

                Assert.Equal(
                    CoreStrings.ExecutionStrategyExistingTransaction(mockExecutionStrategy.GetType().Name, "DbContext.Database.CreateExecutionStrategy()"),
                    Assert.Throws<InvalidOperationException>(
                            () => execute(mockExecutionStrategy))
                        .Message);
            }
        }

        [Fact]
        public void Execute_Action_does_not_throw_when_invoked_twice()
        {
            Execute_does_not_throw_when_invoked_twice((e, f) => e.Execute(() => f()));
        }

        [Fact]
        public void Execute_Func_does_not_throw_when_invoked_twice()
        {
            Execute_does_not_throw_when_invoked_twice((e, f) => e.Execute(f));
        }

        private void Execute_does_not_throw_when_invoked_twice(Action<ExecutionStrategy, Func<int>> execute)
        {
            var executed = false;

            var executionStrategyMock = new TestExecutionStrategy(
                Context,
                shouldRetryOn: e => e is ArgumentOutOfRangeException);

            for (var i = 0; i < 2; i++)
            {
                execute(
                    executionStrategyMock, () =>
                    {
                        if (!executed)
                        {
                            executed = true;
                            throw new ArgumentOutOfRangeException();
                        }

                        return 0;
                    });

                Assert.True(executed);
                executed = false;
            }
        }

        [Fact]
        public void Execute_Action_doesnt_retry_if_succesful()
        {
            Execute_doesnt_retry_if_succesful((e, f) => e.Execute(() => f()));
        }

        [Fact]
        public void Execute_Func_doesnt_retry_if_succesful()
        {
            Execute_doesnt_retry_if_succesful((e, f) => e.Execute(f));
        }

        private void Execute_doesnt_retry_if_succesful(Action<ExecutionStrategy, Func<int>> execute)
        {
            var executionCount = 0;
            execute(CreateFailOnRetryStrategy(), () => executionCount++);

            Assert.Equal(1, executionCount);
        }

        [Fact]
        public void Execute_Action_doesnt_retry_if_suspended()
        {
            Execute_doesnt_retry_if_suspended((e, f) => e.Execute(() => f()));
        }

        [Fact]
        public void Execute_Func_doesnt_retry_if_suspended()
        {
            Execute_doesnt_retry_if_suspended((e, f) => e.Execute(f));
        }

        private void Execute_doesnt_retry_if_suspended(Action<ExecutionStrategy, Func<int>> execute)
        {
            TestExecutionStrategy.Suspended = true;
            var executionCount = 0;
            Assert.Throws<DbUpdateException>(
                () =>
                    execute(
                        CreateFailOnRetryStrategy(), () =>
                        {
                            executionCount++;
                            throw new DbUpdateException("", new ArgumentOutOfRangeException());
                        }));
            TestExecutionStrategy.Suspended = false;

            Assert.Equal(1, executionCount);
        }

        [Fact]
        public void Execute_Action_retries_until_succesful()
        {
            Execute_retries_until_succesful((e, f) => e.Execute(() => f()));
        }

        [Fact]
        public void Execute_Func_retries_until_succesful()
        {
            Execute_retries_until_succesful((e, f) => e.Execute(f));
        }

        private void Execute_retries_until_succesful(Action<ExecutionStrategy, Func<int>> execute)
        {
            var executionStrategyMock = new TestExecutionStrategy(
                Context,
                shouldRetryOn: e => e is ArgumentOutOfRangeException,
                getNextDelay: e => TimeSpan.FromTicks(0));

            var executionCount = 0;

            execute(
                executionStrategyMock, () =>
                {
                    if (executionCount++ < 3)
                    {
                        throw new DbUpdateException("", new ArgumentOutOfRangeException());
                    }

                    return executionCount;
                });

            Assert.Equal(4, executionCount);
        }

        [Fact]
        public void Execute_Action_retries_until_not_retrieable_exception_is_thrown()
        {
            Execute_retries_until_not_retrieable_exception_is_thrown((e, f) => e.Execute(() => f()));
        }

        [Fact]
        public void Execute_Func_retries_until_not_retrieable_exception_is_thrown()
        {
            Execute_retries_until_not_retrieable_exception_is_thrown((e, f) => e.Execute(f));
        }

        private void Execute_retries_until_not_retrieable_exception_is_thrown(Action<ExecutionStrategy, Func<int>> execute)
        {
            var executionStrategyMock = new TestExecutionStrategy(
                Context,
                shouldRetryOn: e => e is ArgumentOutOfRangeException,
                getNextDelay: e => TimeSpan.FromTicks(0));

            var executionCount = 0;

            Assert.Throws<ArgumentNullException>(
                () =>
                    execute(
                        executionStrategyMock, () =>
                        {
                            if (executionCount++ < 3)
                            {
                                throw new ArgumentOutOfRangeException();
                            }

                            throw new ArgumentNullException();
                        }));

            Assert.Equal(4, executionCount);
        }

        [Fact]
        public void Execute_Action_retries_until_limit_is_reached()
        {
            Execute_retries_until_limit_is_reached((e, f) => e.Execute(() => f()));
        }

        [Fact]
        public void Execute_Func_retries_until_limit_is_reached()
        {
            Execute_retries_until_limit_is_reached((e, f) => e.Execute(f));
        }

        private void Execute_retries_until_limit_is_reached(Action<ExecutionStrategy, Func<int>> execute)
        {
            var executionCount = 0;

            var executionStrategyMock = new TestExecutionStrategy(
                Context,
                retryCount: 2,
                shouldRetryOn: e => e is ArgumentOutOfRangeException,
                getNextDelay: e => TimeSpan.FromTicks(0));

            Assert.IsType<ArgumentOutOfRangeException>(
                Assert.Throws<RetryLimitExceededException>(
                        () =>
                            execute(
                                executionStrategyMock, () =>
                                {
                                    if (executionCount++ < 3)
                                    {
                                        throw new ArgumentOutOfRangeException();
                                    }

                                    Assert.True(false);
                                    return 0;
                                }))
                    .InnerException);

            Assert.Equal(3, executionCount);
        }

        [Fact]
        public Task ExecuteAsync_Action_throws_for_an_existing_transaction()
        {
            return ExecuteAsync_throws_for_an_existing_transaction(e => e.ExecuteAsync(() => (Task)Task.FromResult(1)));
        }

        [Fact]
        public Task ExecuteAsync_Func_throws_for_an_existing_transaction()
        {
            return ExecuteAsync_throws_for_an_existing_transaction(e => e.ExecuteAsync(ct => Task.FromResult(1), CancellationToken.None));
        }

        private async Task ExecuteAsync_throws_for_an_existing_transaction(Func<ExecutionStrategy, Task> executeAsync)
        {
            var mockExecutionStrategy = new TestExecutionStrategy(Context);
            using (Context.Database.BeginTransaction())
            {
                Assert.Equal(
                    CoreStrings.ExecutionStrategyExistingTransaction(mockExecutionStrategy.GetType().Name, "DbContext.Database.CreateExecutionStrategy()"),
                    (await Assert.ThrowsAsync<InvalidOperationException>(
                        () => executeAsync(mockExecutionStrategy))).Message);
            }
        }

        [Fact]
        public void ExecuteAsync_Action_throws_for_an_ambient_transaction()
        {
            ExecuteAsync_throws_for_an_ambient_transaction(e => e.ExecuteAsync(() => (Task)Task.FromResult(1)));
        }

        [Fact]
        public void ExecuteAsync_Func_throws_for_an_ambient_transaction()
        {
            ExecuteAsync_throws_for_an_ambient_transaction(e => e.ExecuteAsync(ct => Task.FromResult(1), CancellationToken.None));
        }

        private async void ExecuteAsync_throws_for_an_ambient_transaction(Func<ExecutionStrategy, Task> executeAsync)
        {
            var mockExecutionStrategy = new TestExecutionStrategy(Context);
            using (new TransactionScope(TransactionScopeAsyncFlowOption.Enabled))
            {
                Assert.Equal(
                    CoreStrings.ExecutionStrategyExistingTransaction(mockExecutionStrategy.GetType().Name, "DbContext.Database.CreateExecutionStrategy()"),
                    (await Assert.ThrowsAsync<InvalidOperationException>(
                        () => executeAsync(mockExecutionStrategy)))
                    .Message);
            }
        }

        [Fact]
        public void ExecuteAsync_Action_throws_for_an_enlisted_transaction()
        {
            ExecuteAsync_throws_for_an_enlisted_transaction(e => e.ExecuteAsync(() => (Task)Task.FromResult(1)));
        }

        [Fact]
        public void ExecuteAsync_Func_throws_for_an_enlisted_transaction()
        {
            ExecuteAsync_throws_for_an_enlisted_transaction(e => e.ExecuteAsync(ct => Task.FromResult(1), CancellationToken.None));
        }

        private async void ExecuteAsync_throws_for_an_enlisted_transaction(Func<ExecutionStrategy, Task> executeAsync)
        {
            var mockExecutionStrategy = new TestExecutionStrategy(Context);
            using (var t = new CommittableTransaction())
            {
                Context.Database.EnlistTransaction(t);

                Assert.Equal(
                    CoreStrings.ExecutionStrategyExistingTransaction(mockExecutionStrategy.GetType().Name, "DbContext.Database.CreateExecutionStrategy()"),
                    (await Assert.ThrowsAsync<InvalidOperationException>(
                        () => executeAsync(mockExecutionStrategy)))
                    .Message);
            }
        }

        [Fact]
        public Task ExecuteAsync_Action_does_not_throw_when_invoked_twice()
        {
            return ExecuteAsync_does_not_throw_when_invoked_twice((e, f) => e.ExecuteAsync(() => (Task)f(CancellationToken.None)));
        }

        [Fact]
        public Task ExecuteAsync_Func_does_not_throw_when_invoked_twice()
        {
            return ExecuteAsync_does_not_throw_when_invoked_twice((e, f) => e.ExecuteAsync(f, CancellationToken.None));
        }

        private async Task ExecuteAsync_does_not_throw_when_invoked_twice(Func<ExecutionStrategy, Func<CancellationToken, Task<int>>, Task> executeAsync)
        {
            var executed = false;

            var executionStrategyMock = new TestExecutionStrategy(
                Context,
                shouldRetryOn: e => e is ArgumentOutOfRangeException);

            for (var i = 0; i < 2; i++)
            {
                await executeAsync(
                    executionStrategyMock, ct =>
                    {
                        if (!executed)
                        {
                            executed = true;
                            throw new ArgumentOutOfRangeException();
                        }

                        return Task.FromResult(0);
                    });

                Assert.True(executed);
                executed = false;
            }
        }

        [Fact]
        public Task ExecuteAsync_Action_doesnt_retry_if_succesful()
        {
            return ExecuteAsync_doesnt_retry_if_succesful((e, f) => e.ExecuteAsync(ct => (Task)f(ct), CancellationToken.None));
        }

        [Fact]
        public Task ExecuteAsync_Func_doesnt_retry_if_succesful()
        {
            return ExecuteAsync_doesnt_retry_if_succesful((e, f) => e.ExecuteAsync(f, CancellationToken.None));
        }

        private async Task ExecuteAsync_doesnt_retry_if_succesful(Func<ExecutionStrategy, Func<CancellationToken, Task<int>>, Task> executeAsync)
        {
            var executionCount = 0;
            await executeAsync(CreateFailOnRetryStrategy(), ct => Task.FromResult(executionCount++));

            Assert.Equal(1, executionCount);
        }

        [Fact]
        public Task ExecuteAsync_Action_doesnt_retry_if_suspended()
        {
            return ExecuteAsync_doesnt_retry_if_suspended((e, f) => e.ExecuteAsync(() => (Task)f(CancellationToken.None)));
        }

        [Fact]
        public Task ExecuteAsync_Func_doesnt_retry_if_suspended()
        {
            return ExecuteAsync_doesnt_retry_if_suspended((e, f) => e.ExecuteAsync(f, CancellationToken.None));
        }

        private async Task ExecuteAsync_doesnt_retry_if_suspended(Func<ExecutionStrategy, Func<CancellationToken, Task<int>>, Task> executeAsync)
        {
            TestExecutionStrategy.Suspended = true;
            var executionCount = 0;
            await Assert.ThrowsAsync<DbUpdateException>(
                () =>
                    executeAsync(
                        CreateFailOnRetryStrategy(), ct =>
                        {
                            executionCount++;
                            throw new DbUpdateException("", new ArgumentOutOfRangeException());
                        }));
            TestExecutionStrategy.Suspended = false;

            Assert.Equal(1, executionCount);
        }

        [Fact]
        public Task ExecuteAsync_Action_retries_until_succesful()
        {
            return ExecuteAsync_retries_until_succesful((e, f) => e.ExecuteAsync(ct => (Task)f(ct), CancellationToken.None));
        }

        [Fact]
        public Task ExecuteAsync_Func_retries_until_succesful()
        {
            return ExecuteAsync_retries_until_succesful((e, f) => e.ExecuteAsync(f, CancellationToken.None));
        }

        private async Task ExecuteAsync_retries_until_succesful(Func<ExecutionStrategy, Func<CancellationToken, Task<int>>, Task> executeAsync)
        {
            var executionStrategyMock = new TestExecutionStrategy(
                Context,
                shouldRetryOn: e => e is ArgumentOutOfRangeException,
                getNextDelay: e => TimeSpan.FromTicks(0));

            var executionCount = 0;

            await executeAsync(
                executionStrategyMock, ct =>
                {
                    if (executionCount++ < 3)
                    {
                        throw new DbUpdateException("", new ArgumentOutOfRangeException());
                    }

                    return Task.FromResult(executionCount);
                });

            Assert.Equal(4, executionCount);
        }

        [Fact]
        public Task ExecuteAsync_Action_retries_until_not_retrieable_exception_is_thrown()
        {
            return ExecuteAsync_retries_until_not_retrieable_exception_is_thrown(
                (e, f) => e.ExecuteAsync(ct => (Task)f(ct), CancellationToken.None));
        }

        [Fact]
        public Task ExecuteAsync_Func_retries_until_not_retrieable_exception_is_thrown()
        {
            return ExecuteAsync_retries_until_not_retrieable_exception_is_thrown((e, f) => e.ExecuteAsync(f, CancellationToken.None));
        }

        private async Task ExecuteAsync_retries_until_not_retrieable_exception_is_thrown(
            Func<ExecutionStrategy, Func<CancellationToken, Task<int>>, Task> executeAsync)
        {
            var executionStrategyMock = new TestExecutionStrategy(
                Context,
                shouldRetryOn: e => e is ArgumentOutOfRangeException,
                getNextDelay: e => TimeSpan.FromTicks(0));

            var executionCount = 0;

            await Assert.ThrowsAsync<ArgumentNullException>(
                () => executeAsync(
                    executionStrategyMock, ct =>
                    {
                        if (executionCount++ < 3)
                        {
                            throw new ArgumentOutOfRangeException();
                        }

                        throw new ArgumentNullException();
                    }));

            Assert.Equal(4, executionCount);
        }

        [Fact]
        public Task ExecuteAsync_Action_retries_until_limit_is_reached()
        {
            return ExecuteAsync_retries_until_limit_is_reached((e, f) => e.ExecuteAsync(ct => (Task)f(ct), CancellationToken.None));
        }

        [Fact]
        public Task ExecuteAsync_Func_retries_until_limit_is_reached()
        {
            return ExecuteAsync_retries_until_limit_is_reached((e, f) => e.ExecuteAsync(f, CancellationToken.None));
        }

        private async Task ExecuteAsync_retries_until_limit_is_reached(Func<ExecutionStrategy, Func<CancellationToken, Task<int>>, Task> executeAsync)
        {
            var executionCount = 0;

            var executionStrategyMock = new TestExecutionStrategy(
                Context,
                retryCount: 2,
                shouldRetryOn: e => e is ArgumentOutOfRangeException,
                getNextDelay: e => TimeSpan.FromTicks(0));

            // ReSharper disable once PossibleNullReferenceException
            Assert.IsType<ArgumentOutOfRangeException>(
                (await Assert.ThrowsAsync<RetryLimitExceededException>(
                    () =>
                        executeAsync(
                            executionStrategyMock, ct =>
                            {
                                if (executionCount++ < 3)
                                {
                                    throw new DbUpdateException("", new ArgumentOutOfRangeException());
                                }

                                Assert.True(false);
                                return Task.FromResult(0);
                            }))).InnerException.InnerException);

            Assert.Equal(3, executionCount);
        }

        protected DbContext CreateContext()
            => InMemoryTestHelpers.Instance.CreateContext(
                InMemoryTestHelpers.Instance.CreateServiceProvider(
                    new ServiceCollection()
                        .AddScoped<IDbContextTransactionManager, TestInMemoryTransactionManager>()),
                InMemoryTestHelpers.Instance.CreateOptions());

        private class TestExecutionStrategy : ExecutionStrategy
        {
            private readonly Func<Exception, bool> _shouldRetryOn;
            private readonly Func<Exception, TimeSpan?> _getNextDelay;

            public TestExecutionStrategy(
                DbContext context,
                int? retryCount = null,
                Func<Exception, bool> shouldRetryOn = null,
                Func<Exception, TimeSpan?> getNextDelay = null)
                : base(
                    context,
                    retryCount ?? DefaultMaxRetryCount,
                    DefaultMaxDelay)
            {
                _shouldRetryOn = shouldRetryOn;
                _getNextDelay = getNextDelay;
            }

            protected internal override bool ShouldRetryOn(Exception exception)
                => _shouldRetryOn?.Invoke(exception) == true;

            protected override TimeSpan? GetNextDelay(Exception lastException)
            {
                var baseDelay = base.GetNextDelay(lastException);
                return baseDelay != null && _getNextDelay != null ? _getNextDelay.Invoke(lastException) : baseDelay;
            }

            public TimeSpan? GetNextDelayBase(Exception lastException)
            {
                ExceptionsEncountered.Add(lastException);
                return base.GetNextDelay(lastException);
            }

            public static new bool Suspended
            {
                set => ExecutionStrategy.Suspended = value;
            }
        }
    }
}
